from baddns.base import BadDNS_base
from baddns.lib.dnsmanager import DNSManager
from baddns.lib.httpmanager import HttpManager
from baddns.modules.cname import BadDNS_cname
from baddns.lib.findings import Finding

import logging

log = logging.getLogger(__name__)


class BadDNS_txt(BadDNS_base):
    name = "TXT"
    description = "Check TXT record contents for hijackable domains"

    def __init__(self, target, **kwargs):
        super().__init__(target, **kwargs)
        self.target = target
        self.target_dnsmanager = DNSManager(
            target, dns_client=self.dns_client, custom_nameservers=self.custom_nameservers
        )
        self.target_httpmanager = HttpManager(
            self.target, http_client_class=self.http_client_class, skip_redirects=True
        )
        self.cname_findings = None
        self.cname_findings_direct = None
        self.reference_data = {}

    async def dispatch(self):
        self.cname_findings_direct = []
        self.cname_findings = []

        await self.target_dnsmanager.dispatchDNS(omit_types=["A", "AAAA", "CNAME", "NS", "SOA", "MX", "NSEC"])
        if self.target_dnsmanager.answers["TXT"] == None:
            log.debug("No TXT records found, aborting")
            return False

        for txt_record in self.target_dnsmanager.answers["TXT"]:
            log.debug(f"Got TXT record [{txt_record}]")

            for match in DNSManager.dns_name_regex.finditer(txt_record):
                start, end = match.span()
                host = txt_record[start:end]
                self.infomsg(f"Found host [{host}] in TXT record [{txt_record}] and analyzing with CNAME module")

                cname_instance_direct = BadDNS_cname(
                    host,
                    custom_nameservers=self.custom_nameservers,
                    signatures=self.signatures,
                    direct_mode=True,
                    parent_class="txt",
                    http_client_class=self.http_client_class,
                    dns_client=self.dns_client,
                )
                if await cname_instance_direct.dispatch():
                    self.cname_findings_direct.append(
                        {
                            "finding": cname_instance_direct.analyze(),
                            "description": "Vulnerable Host in TXT Record",
                            "trigger": self.target_dnsmanager.target,
                        }
                    )

                cname_instance = BadDNS_cname(
                    host,
                    custom_nameservers=self.custom_nameservers,
                    signatures=self.signatures,
                    direct_mode=False,
                    parent_class="txt",
                    http_client_class=self.http_client_class,
                    dns_client=self.dns_client,
                )
                if await cname_instance.dispatch():
                    self.cname_findings.append(
                        {
                            "finding": cname_instance.analyze(),
                            "description": "Vulnerable Host in TXT Record",
                            "trigger": self.target_dnsmanager.target,
                        }
                    )

        return True

    def _convert_findings(self, finding_sets):
        converted_findings = []
        for finding_set in finding_sets:
            for finding in finding_set["finding"]:
                finding_dict = finding.to_dict()
                log.debug(f"Found finding during cname check for target: {finding_dict['target']}")
                converted_findings.append(
                    Finding(
                        {
                            "target": self.target,
                            "description": f"{finding_set['description']}. Original Event: [{finding_dict['description']}]",
                            "confidence": finding_dict["confidence"],
                            "signature": finding_dict["signature"],
                            "indicator": finding_dict["indicator"],
                            "trigger": finding_set["trigger"],
                            "module": type(self),
                        }
                    )
                )
        return converted_findings

    def analyze(self):
        findings = []
        log.debug("in txt analyze")
        if self.cname_findings_direct:
            findings.extend(self._convert_findings(self.cname_findings_direct))
        if self.cname_findings:
            findings.extend(self._convert_findings(self.cname_findings))
        return findings



import logging

log = logging.getLogger(__name__)


class BadDNS_base:
    def __init__(
        self,
        target,
        http_client_class=None,
        dns_client=None,
        signatures=None,
        custom_nameservers=None,
        cli=False,
        **kwargs,
    ):
        self.target = self.set_target(target)
        self.http_client_class = http_client_class
        self.dns_client = dns_client
        self.signatures = signatures
        self.custom_nameservers = custom_nameservers
        self.parent_class = kwargs.get("parent_class", "self")
        self.cli = cli

    # hook to allow external manipulation of target assignment
    def set_target(self, target):
        return target

    def infomsg(self, msg):
        if self.cli:
            log.info(msg)
        else:
            log.debug(msg)


def get_all_modules(*args, **kwargs):
    return [m for m in BadDNS_base.__subclasses__()]



import sys


# Temporary workaround for: https://github.com/blacklanternsecurity/baddns/issues/402
class noop:
    def __getattr__(self, item):
        def method(*args, **kwargs):
            pass  # Method does nothing

        return method


sys.modules["imp"] = noop()

import os
from contextlib import contextmanager


# Another temporary workaround until https://github.com/richardpenman/whois gets updated version pushed to pypi :( :( :(
@contextmanager
def suppress_stdout():
    original_stdout = sys.stdout
    sys.stdout = open(os.devnull, "w")
    try:
        yield
    finally:
        sys.stdout = original_stdout


import whois
import logging
import asyncio
import tldextract
from datetime import date, datetime, timedelta
from dateutil import parser as date_parser


log = logging.getLogger(__name__)


class WhoisManager:
    def __init__(self, target):
        self.target = target
        self.whois_result = None

    async def dispatchWHOIS(self):
        ext = tldextract.extract(self.target)
        log.debug(f"Extracted base domain [{ext.registered_domain}] from [{self.target}]")
        log.debug(f"Submitting WHOIS query for {ext.registered_domain}")
        try:
            with suppress_stdout():
                w = await asyncio.to_thread(whois.whois, ext.registered_domain)
            log.debug(f"Got response to whois request for {ext.registered_domain}")
            self.whois_result = {"type": "response", "data": w}
        except whois.parser.PywhoisError as e:
            log.debug(f"Got PywhoisError for whois request for {ext.registered_domain}")
            self.whois_result = {"type": "error", "data": str(e)}

    def analyzeWHOIS(self):
        if self.whois_result:
            whois_findings = []
            if self.whois_result["type"] == "error":
                log.debug("whois result was an error")
                if "No match for" in self.whois_result["data"]:
                    whois_findings.append("unregistered")

            elif self.whois_result["type"] == "response":
                log.debug("whois resulted in a response")
                expiration_data = self.whois_result.get("data", {}).get("expiration_date", None)

                if isinstance(expiration_data, list):
                    log.debug("Expiration data:")
                    log.debug(expiration_data)
                    log.debug("Got multiple expiration dates. Falling back to the latest...")
                    expiration_date = max(expiration_data)
                else:
                    expiration_date = expiration_data

                expiration_date = self.date_parse(expiration_date)
                if expiration_date:
                    current_date = date.today()
                    expiration_plus_one = expiration_date.date() + timedelta(days=1)
                    if expiration_plus_one < current_date:
                        log.debug(
                            f"Current Date (minus one) ({current_date.strftime('%Y-%m-%d')}) is after Expiration Date ({expiration_date.date().strftime('%Y-%m-%d')})"
                        )
                        whois_findings.append(
                            f"Registration Expired (Expiration: [{expiration_date.strftime('%Y-%m-%d %H:%M:%S')}]"
                        )
                    else:
                        log.debug(f"Domain {self.target} is not expired")
            return whois_findings
        else:
            log.debug("whois_result was NoneType")

    @staticmethod
    def date_parse(unknown_date):
        # Check if it's already a datetime object
        if isinstance(unknown_date, datetime):
            return unknown_date

        # If it's a string, try to parse it
        if isinstance(unknown_date, str):
            try:
                return date_parser.parse(unknown_date)
            except ValueError as e:
                log.debug(f"Failed to parse date from string: {unknown_date}. Error: {e}")
                return None

        log.debug(f"Unsupported date object type: {type(unknown_date)}. Value: {unknown_date}")
        return None



#!/usr/bin/env python3
# baddns
# Black Lantern Security - https://www.blacklanternsecurity.com
# @paulmmueller

import re
import sys
import asyncio
import argparse
import logging
import pkg_resources

from .lib.errors import BadDNSSignatureException, BadDNSCLIException
from .lib.logging import setup_logging
from .lib.loader import load_signatures

from baddns.base import get_all_modules


from colorama import Fore, Style, init

init(autoreset=True)  # Automatically reset the color to default after each print statement

modules = get_all_modules()


class CustomArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        self.print_usage()
        log.error(message)
        raise BadDNSCLIException(message)


def print_version():
    version = pkg_resources.get_distribution("baddns").version
    if version == "1.0.0":
        version = "Unknown (Running w/poetry?)"
    print(f"Version - {version}\n")


def validate_target(
    arg_value, pattern=re.compile(r"^(?:[a-z0-9_](?:[a-z0-9-_]{0,61}[a-z0-9])?\.)+[a-z0-9][a-z0-9-]{0,61}[a-z0-9]$")
):
    if not pattern.match(arg_value):
        raise argparse.ArgumentTypeError("Target subdomain is not correctly formatted")
    return arg_value


def validate_nameservers(
    arg_value,
    pattern=re.compile(
        r"^((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)(,((25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?))*$"
    ),
):
    if not pattern.match(arg_value):
        raise argparse.ArgumentTypeError("Nameservers argument is incorrectly formatted")
    return arg_value


def validate_modules(arg_value, pattern=re.compile(r"^[a-zA-Z0-9_]+(,[a-zA-Z0-9_]+)*$")):
    if not pattern.match(arg_value):
        raise argparse.ArgumentTypeError(
            "The format of provided modules is incorrect. Use comma-separated values without spaces."
        )

    modules_provided = [m.upper() for m in arg_value.split(",")]
    for m in modules_provided:
        if not any(m in module.name.upper() for module in modules):
            raise argparse.ArgumentTypeError(
                f"'{m}' is not a recognized module. Please check the module name or use '-l' to list available modules."
            )
    return arg_value


async def execute_module(ModuleClass, target, custom_nameservers, signatures):
    findings = None
    try:
        module_instance = ModuleClass(target, custom_nameservers=custom_nameservers, signatures=signatures, cli=True)
    except BadDNSSignatureException as e:
        log.error(f"Error loading signatures: {e}")
        raise BadDNSCLIException(f"Error loading signatures: {e}")

    log.info(f"Starting [{module_instance.name}] module with target [{target}]")
    if await module_instance.dispatch():
        findings = module_instance.analyze()
        if findings:
            print(f"{Fore.GREEN}{'Vulnerable!'}{Style.RESET_ALL}")
            for finding in findings:
                print(finding.to_dict())
    return findings


async def _main():
    setup_logging()
    global log
    log = logging.getLogger("baddns")

    parser = CustomArgumentParser(description="Check subdomains for subdomain takeovers and other DNS tomfoolery")
    print(f"{Fore.GREEN}{ascii_art_banner}{Style.RESET_ALL}")
    print_version()

    parser.add_argument(
        "-n",
        "--custom-nameservers",
        type=validate_nameservers,
        help="Provide a list of custom nameservers separated by comma.",
    )

    parser.add_argument(
        "-c",
        "--custom-signatures",
        help="Use an alternate directory for loading signatures",
    )

    parser.add_argument(
        "-l", "--list-modules", action="store_true", help="List available modules and their descriptions."
    )

    parser.add_argument(
        "-m",
        "--modules",
        type=validate_modules,
        help="Comma separated list of module names to use. Ex: module1,module2,module3",
    )

    parser.add_argument("-d", "--debug", action="store_true", help="Enable debug logging")

    parser.add_argument("target", nargs="?", type=validate_target, help="subdomain to analyze")
    args = parser.parse_args()

    if not args.target and not args.list_modules:
        parser.error("the following arguments are required: target")

    if args.list_modules:
        r = get_all_modules()
        print("Available Modules:")
        for m in r:
            log.info(f"[{m.name}] - {m.description}")
        sys.exit(0)

    if args.debug:
        log.setLevel(logging.DEBUG)

    # Get all available modules
    all_modules = get_all_modules()

    # If the user provided the -m or --modules argument, filter the modules accordingly
    if args.modules:
        included_module_names = [name.strip().upper() for name in args.modules.split(",")]
        modules_to_execute = [module for module in all_modules if module.name.upper() in included_module_names]
    else:
        modules_to_execute = all_modules  # Default to all modules if -m is not provided
        log.info(
            f"Running with all modules [{', '.join([module.name for module in modules_to_execute])}] (-m to specify)"
        )

    custom_signatures = None
    if args.custom_signatures:
        custom_signatures = args.custom_signatures
        log.info(f"Using custom signatures directory: [{args.custom_signatures}]")

    custom_nameservers = None
    if args.custom_nameservers:
        custom_nameservers = args.custom_nameservers.split(",")
        log.info(f"Using custom nameservers: [{', '.join(custom_nameservers)}]")

    signatures = load_signatures(signatures_dir=custom_signatures)

    for ModuleClass in modules_to_execute:
        await execute_module(ModuleClass, args.target, custom_nameservers, signatures)


def main():
    try:
        asyncio.run(_main())
    except asyncio.CancelledError:
        log.error("Got asyncio.CancelledError")

    except BadDNSCLIException:
        sys.exit(1)

    except KeyboardInterrupt:
        sys.exit(1)


ascii_art_banner = """
  __ )              |      |              
  __ \    _` |   _` |   _` |  __ \    __| 
  |   |  (   |  (   |  (   |  |   | \__ \ 
 ____/  \__,_| \__,_| \__,_| _|  _| ____/ 
                                          
"""


if __name__ == "__main__":
    main()



from baddns.base import BadDNS_base

from baddns.lib.dnsmanager import DNSManager
from baddns.lib.dnswalk import DnsWalk
from baddns.lib.findings import Finding

import logging

log = logging.getLogger(__name__)


class BadDNS_ns(BadDNS_base):
    name = "NS"
    description = "Check for dangling NS records, and interrogate them for takeover opportunities"

    def __init__(self, target, **kwargs):
        super().__init__(target, **kwargs)
        self.target_dnsmanager = DNSManager(
            target, dns_client=self.dns_client, custom_nameservers=self.custom_nameservers
        )

    async def dispatch(self):
        # omit everything except CNAME. If there is a CNAME chain, we want to run against the end of it.
        await self.target_dnsmanager.dispatchDNS(omit_types=[["A", "AAAA", "MX", "NS", "SOA", "TXT", "NSEC"]])

        if self.target_dnsmanager.answers["CNAME"] != None:
            self.infomsg(
                f"Detected CNAME(S). Will set target to end of CNAME chain: [{self.target_dnsmanager.answers['CNAME'][-1]}]"
            )
            self.target_dnsmanager.target = self.target_dnsmanager.answers["CNAME"][-1]
            self.target = self.target_dnsmanager.answers["CNAME"][-1]
            self.target_dnsmanager.reset_answers()

        await self.target_dnsmanager.dispatchDNS(omit_types=["CNAME", "NS"])

        dnswalk = DnsWalk(self.target_dnsmanager)
        self.target_dnsmanager.answers["NS"] = await dnswalk.ns_trace(self.target)
        return True

    @staticmethod
    def get_substring_matches(nameservers, strings):
        matched_nameservers = set()
        matched_signatures = set()

        for ns in nameservers:
            for s in strings:
                if s in ns:
                    matched_nameservers.add(ns)
                    matched_signatures.add(s)

        if not matched_nameservers and not matched_signatures:
            return None

        return list(matched_nameservers), list(matched_signatures)

    def analyze(self):
        log.debug("Staring analysis")
        findings = []
        if self.target_dnsmanager.answers["NS"] and len(self.target_dnsmanager.answers["NS"]) > 0:
            target_nameservers = self.target_dnsmanager.answers["NS"]
            log.debug("Nameserver(s) found. Continuing...")
        else:
            return False
        if self.target_dnsmanager.answers["SOA"] == None:
            log.debug("No SOA record found w/nameservers present")
            r = None
            for sig in self.signatures:
                if sig.signature["mode"] == "dns_nosoa":
                    sig_nameservers = [ns for ns in sig.signature["identifiers"]["nameservers"]]
                    r = self.get_substring_matches(target_nameservers, sig_nameservers)
                    if r:
                        findings.append(
                            Finding(
                                {
                                    "target": self.target_dnsmanager.target,
                                    "description": "Dangling NS Records (NS records without SOA) with known impact",
                                    "confidence": "PROBABLE",
                                    "signature": sig.signature["service_name"],
                                    "indicator": f"DnsWalk Analysis with signature match: {r[1]}",
                                    "trigger": target_nameservers,
                                    "module": type(self),
                                }
                            )
                        )
                        log.debug(
                            f"Found match for for target nameservers {', '.join(target_nameservers)} with signature [{sig.signature['service_name']}]"
                        )
                        return findings
            log.debug(
                f"No signature found, falling back to report generic dangling NS record for nameservers: [{', '.join(target_nameservers)}]]"
            )
            findings.append(
                Finding(
                    {
                        "target": self.target_dnsmanager.target,
                        "description": "Dangling NS Records (NS records without SOA)",
                        "confidence": "POSSIBLE",
                        "signature": "N/A",
                        "indicator": "DNSWalk Analysis",
                        "trigger": target_nameservers,
                        "module": type(self),
                    }
                )
            )

        return findings



from baddns.base import BadDNS_base

from baddns.lib.findings import Finding
from baddns.lib.dnsmanager import DNSManager
from baddns.lib.whoismanager import WhoisManager
import logging

log = logging.getLogger(__name__)


class BadDNS_mx(BadDNS_base):
    name = "MX"
    description = "Check for dangling MX records and assess their base domains for availability"

    def __init__(self, target, **kwargs):
        super().__init__(target, **kwargs)

        self.target_dnsmanager = DNSManager(
            target, dns_client=self.dns_client, custom_nameservers=self.custom_nameservers
        )
        self.mx_whoismanager = {}

    async def dispatch(self):
        await self.target_dnsmanager.dispatchDNS(omit_types=["A", "AAAA", "CNAME", "NS", "SOA", "TXT", "NSEC"])
        if self.target_dnsmanager.answers["MX"] == None:
            log.debug("No MX records found, aborting")
            return False

        for mx_record in self.target_dnsmanager.answers["MX"]:
            log.debug(f"performing WHOIS lookup for [{mx_record}]")
            self.mx_whoismanager[mx_record] = WhoisManager(mx_record)
            await self.mx_whoismanager[mx_record].dispatchWHOIS()
            log.debug(f"WHOIS dispatch [{mx_record}] complete")
        return True

    def analyze(self):
        findings = []
        log.debug("Received the following MX answers:")
        log.debug(self.target_dnsmanager.answers["MX"])

        for whois_domain, whois_data in self.mx_whoismanager.items():
            for whois_finding in whois_data.analyzeWHOIS():
                findings.append(
                    Finding(
                        {
                            "target": self.target_dnsmanager.target,
                            "description": f"MX {whois_finding}",
                            "confidence": "CONFIRMED",
                            "signature": "N/A",
                            "indicator": "Whois Data",
                            "trigger": whois_domain,
                            "module": type(self),
                        }
                    )
                )

        return findings



import os
import yaml
import logging
import pkg_resources

log = logging.getLogger(__name__)

from .errors import BadDNSSignatureException
from .signature import BadDNSSignature


def load_signatures(signatures_dir=None):
    signatures = []
    if signatures_dir:
        if not os.path.exists(signatures_dir):
            raise BadDNSSignatureException(f"Signatures directory [{signatures_dir}] does not exist")
    else:
        signatures_dir = pkg_resources.resource_filename("baddns", "signatures")

    log.debug(f"attempting to load signatures from: {signatures_dir}")

    for filename in os.listdir(signatures_dir):
        if filename.endswith(".yml"):
            file_path = os.path.join(signatures_dir, filename)

            # Open each file and load the YAML contents
            try:
                with open(file_path, "r") as file:
                    signature_data = yaml.safe_load(file)
                    signature = BadDNSSignature()
                    signature.initialize(**signature_data)
                    signatures.append(signature)
            except BadDNSSignatureException as e:
                log.error(f"Error loading signature from {filename}: {e}")
    if len(signatures) == 0:
        raise BadDNSSignatureException(f"No signatures were successfuly loaded from [{signatures_dir}]")
    else:
        log.debug(f"Loaded [{str(len(signatures))}] signatures from [{signatures_dir}]")

    return signatures



import re
import os
import sys
import httpx
import yaml
import logging

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.dirname(SCRIPT_DIR))

# from lib.errors import BadDNSMatcherException

log = logging.getLogger(__name__)


class Matcher:
    def __init__(self, rules):
        if isinstance(rules, str):  # YAML input is a string
            try:
                self.rules = yaml.safe_load(rules)
            except yaml.YAMLError as e:
                raise ValueError(f"Error parsing YAML: {e}")
        elif isinstance(rules, dict):  # YAML input is a dict
            self.rules = rules
        else:
            raise TypeError("yaml_rules must be a YAML string or a dict")

    def _status(self, criteria):
        negative = criteria.get("negative", False)
        return (
            self.response.status_code != criteria["status"]
            if negative
            else self.response.status_code == criteria["status"]
        )

    def _word(self, criteria):
        words = criteria["words"]
        part = criteria.get("part", "body").lower()
        negative = criteria.get("negative", False)

        if part == "header":
            text = str(self.response.headers)
        elif part == "body":
            text = self.response.text

        # we can ignore this because are already adding these entries into the identifiers
        elif part == "host":
            return True
        else:
            raise ValueError(f"Unknown part: {part}")

        condition = criteria.get("condition", "and")
        if condition == "and":
            return not all(word in text for word in words) if negative else all(word in text for word in words)
        elif condition == "or":
            return not any(word in text for word in words) if negative else any(word in text for word in words)

    def _regex(self, criteria):
        matches = []
        negative = criteria.get("negative", False)
        for pattern in criteria["regex"]:
            regex = re.compile(pattern)
            if "part" in criteria and criteria["part"].lower() == "header":
                match = any(regex.search(header_value) for header_value in self.response.headers.values())
            else:
                match = bool(regex.search(self.response.text))
            matches.append(match)
        condition = criteria.get("condition", "and")
        if condition == "and":
            return not all(matches) if negative else all(matches)
        elif condition == "or":
            return not any(matches) if negative else any(matches)

    def is_match(self, response):
        if not isinstance(response, httpx.Response):
            raise TypeError("response must be an httpx.Response object")
        self.response = response
        matchers_condition = self.rules.get("matchers-condition", "and")
        results = []
        matcher_rule = self.rules.get("matcher_rule", [])
        for matcher in matcher_rule.get("matchers", []):
            match_type = matcher["type"]
            match_func = getattr(self, f"_{match_type}", None)

            if match_func:
                result = match_func(matcher)
                results.append(result)

        if matchers_condition == "and":
            return all(results)
        elif matchers_condition == "or":
            return any(results)
        return False



class BadDNSException(Exception):
    pass


class BadDNSSignatureException(BadDNSException):
    pass


class BadDNSMatcherException(BadDNSException):
    pass


class BadDNSCLIException(BadDNSException):
    pass


class BadDNSFindingException(BadDNSException):
    pass



import dns.query
import dns.message
import dns.rdatatype
import dns.name
import dns.resolver
import logging
import asyncio
import random

log = logging.getLogger(__name__)


class DnsWalk:
    root_servers = [
        "198.41.0.4",
        "199.9.14.201",
        "192.33.4.12",
        "199.7.91.13",
        "192.203.230.10",
        "192.5.5.241",
        "192.112.36.4",
        "198.97.190.53",
        "192.36.148.17",
        "192.58.128.30",
        "193.0.14.129",
        "199.7.83.42",
        "202.12.27.33",
    ]

    max_depth = 10
    raw_query_max_retries = 6

    def __init__(self, dns_manager):
        self.dns_manager = dns_manager

    async def a_resolve(self, nameserver):
        nameserver_ips = set()
        a_query_results = await self.dns_manager.do_resolve(nameserver, "A")
        if a_query_results:
            for result in a_query_results:
                nameserver_ips.add(result)
            return list(nameserver_ips)
        else:
            return None

    async def ns_trace(self, target):
        log.debug(f"Attempting to find NS records for {target}")
        nameserver_ips = self.root_servers[:]
        solved_nameservers = await self.ns_recursive_solve(nameserver_ips, target, depth=0)
        if solved_nameservers == None:
            solved_nameservers = []
        log.debug(f"Found the following name servers: {solved_nameservers}")
        return solved_nameservers

    async def raw_query_with_retry(self, target, nameserver_ip):
        retries = 0
        while retries < self.raw_query_max_retries:
            try:
                response_msg, used_tcp = await dns.asyncquery.udp_with_fallback(target, nameserver_ip, timeout=6.0)
                log.debug("raw_query_with_retry: Had to fall back to TCP")
                return response_msg, used_tcp
            except (dns.exception.Timeout, dns.message.Truncated) as e:
                log.debug(f"raw_query_with_retry: Query failed due to {e}. Retrying...")
                retries += 1
            except Exception as e:
                log.debug(f"raw_query_with_retry: An unexpected error occurred: {e}. Aborting.")
                retries += 1
            await asyncio.sleep(12)  # Wait before retrying. We don't want to piss off the root DNS servers.
        log.warning("raw_query_with_retry: Max retries reached. Query failed.")
        return None, None

    async def ns_recursive_solve(self, nameserver_ips, target, depth=0):
        if depth > self.max_depth:
            log.debug(
                f"Reached max depth {str(self.max_depth)} attempting to resolve NS records (infinite loop) on target [{target}]"
            )
            return []
        final_results = set()
        log.debug(
            f"Recursive instance: nameserver_ips [{' '.join(nameserver_ips)}] target [{target}] at depth {str(depth)}"
        )
        depth += 1
        next_nameservers = set()

        # randomize which nameservers we ask first to help minimize rate-limiting
        for nameserver_ip in random.sample(list(nameserver_ips), len(nameserver_ips)):
            log.debug(f"Asking nameserver [{nameserver_ip}] NS records on {target}")
            query_msg = dns.message.make_query(target, dns.rdatatype.NS)
            response_msg, used_tcp = await self.raw_query_with_retry(query_msg, nameserver_ip)
            if response_msg != None:
                log.debug(f"Got response message: \n{response_msg}")
                if response_msg.authority:
                    log.debug(f"Server [{nameserver_ip}] responded with authority section")
                    for ns_rrset in response_msg.authority:
                        for rr in ns_rrset:
                            if rr.rdtype == dns.rdatatype.SOA:
                                return None
                            if rr.rdtype == dns.rdatatype.NS:
                                rr_domain = rr.target.to_text().rstrip(".")
                                log.debug(f"Received NS record for [{rr_domain}]")
                                rr_ips = await self.a_resolve(rr_domain)
                                if rr_ips:
                                    log.debug(f"Resolved [{rr_domain}] to ip(s) [{' '.join(rr_ips)}]")
                                    next_nameservers.update(rr_ips)
                                else:
                                    log.debug(f"Could not resolve [{rr_domain}] to an IP!")

                                log.debug(f"Adding {rr_domain} to temp results list, pending deeper results")
                                final_results.add(rr_domain)
                    # If we were provided an authority section but nothing resolved, report the last set - they could be dangling
                    if not next_nameservers:
                        log.debug("None of the servers provded in the authority section resolved")
                        log.debug(f"Submitting [{' '.join(final_results)}] as final result")
                        return list(final_results)

                    # If the next set of nameservers is the same as the set we currently had, continuing would cause an infinite loop. Return an empty list indicating no additional results.
                    if set(nameserver_ips) == set(next_nameservers):
                        log.debug("Got same set of nameservers - preventing infinite loop by returning now")
                        return []

                    # If we had an authority section, and at least one resolved, we need to recurse again
                    log.debug("Resolvable authority results were found. Recursing deeper")
                    recurse = await self.ns_recursive_solve(next_nameservers, target, depth=depth)

                    # If we tried again and didn't get an answer, we might be pointing to a real NS server with no results - these could be dangling
                    if recurse == []:
                        log.debug("Did not get any additional results from latest recursion.")
                        log.debug(f"Submitting [{' '.join(final_results)}] as final result")
                        return list(final_results)
                    else:
                        # If we got results from recursing, we are finishing and are shooting the results back up to the top
                        log.debug(
                            "Results were returned from downstream recurse. Forwarding the results up the stack."
                        )
                        return recurse
                else:
                    log.debug("No Authority section was found. Checking for answers section...")
                    # If there was no authority section, but there was an answer section, we can trust the results
                    if response_msg.answer and len(response_msg.answer) > 0:
                        log.debug("Found answer section!")
                        for rrset in response_msg.answer:
                            for rr in rrset:
                                if rr.rdtype == dns.rdatatype.CNAME:
                                    solved_cname = await self.dns_manager.do_resolve(rr.to_text().rstrip("."), "NS")
                                    if solved_cname:
                                        final_results.update(solved_cname)
                                elif rr.rdtype == dns.rdatatype.NS:
                                    log.debug(f"adding {rr.target.to_text().rstrip('.')} to results list")
                                    final_results.add(rr.target.to_text().rstrip("."))
                    # if there was no authority section, and no answer section, we stop and report whatever the last set was, or and empty set (Probably just has no nameservers)
                    else:
                        log.debug("There was no answer section (or authority section)")
                log.debug("Moving on to next nameserver at current recurse level")
        log.debug("Exhausted entire list of nameservers. Returning any results found.")
        return list(final_results)



from baddns.base import BadDNS_base
from baddns.lib.dnsmanager import DNSManager
from baddns.lib.findings import Finding

import logging
import tldextract

log = logging.getLogger(__name__)


class BadDNS_nsec(BadDNS_base):
    name = "NSEC"
    description = "Enumerate subdomains by NSEC-walking"

    def __init__(self, target, **kwargs):
        super().__init__(target, **kwargs)
        self.target = target
        self.target_dnsmanager = DNSManager(
            target, dns_client=self.dns_client, custom_nameservers=self.custom_nameservers
        )
        self.nsec_chain = []

    async def get_nsec_record(self, domain):
        result = await self.target_dnsmanager.do_resolve(domain, "NSEC")
        if result:
            return result[0].replace("\\000.", "")

    async def nsec_walk(self, domain):
        log.debug("in nsec_walk")
        current_domain = domain
        while 1:
            next_domain = await self.get_nsec_record(current_domain)

            # NSEC wildcard protection
            if current_domain == next_domain and len(self.nsec_chain) == 1:
                return False

            if next_domain is None or next_domain in self.nsec_chain:
                break
            log.debug(f"Found additional NSEC record: {next_domain}")
            if not next_domain.startswith("\\"):
                self.nsec_chain.append(next_domain)
            current_domain = next_domain
        return True

    async def dispatch(self):
        log.debug("in dispatch")
        target_base_domain = tldextract.extract(self.target).registered_domain
        await self.target_dnsmanager.dispatchDNS(omit_types=["A", "AAAA", "CNAME", "NS", "SOA", "MX", "TXT"])
        if self.target_dnsmanager.answers["NSEC"] == None:
            log.debug("No NSEC records found, aborting")
            return False

        self.nsec_chain.append(self.target)
        nsec_walk = await self.nsec_walk(self.target)
        if nsec_walk:
            self.infomsg(f"NSEC Records detected, attempting NSEC walk against domain [{self.target}]")
            self.nsec_chain.remove(self.target)
            nonmatching_results = len(
                [host for host in self.nsec_chain if not host.endswith(f".{target_base_domain}")]
            )
            if len(self.nsec_chain) == nonmatching_results == 1:
                log.debug(
                    f"Aborting because NSEC chain contained only 1 result [{self.nsec_chain[0]}] which did not match the base domain of the target"
                )
                return False
            return True
        return False

    def analyze(self):
        log.debug("in analyze")
        findings = []

        findings.append(
            Finding(
                {
                    "target": self.target_dnsmanager.target,
                    "description": f"DNSSEC NSEC Zone Walking Enabled for domain: [{self.target}]",
                    "confidence": "CONFIRMED",
                    "signature": "N/A",
                    "indicator": "NSEC Records",
                    "trigger": self.target,
                    "module": type(self),
                    "found_domains": self.nsec_chain,
                }
            )
        )
        return findings



import re

from baddns.base import BadDNS_base
from baddns.lib.dnsmanager import DNSManager
from baddns.lib.httpmanager import HttpManager
from baddns.modules.cname import BadDNS_cname
from baddns.lib.findings import Finding

import logging
from urllib.parse import urlparse

log = logging.getLogger(__name__)


class BadDNS_references(BadDNS_base):
    name = "references"
    description = "Check HTML content for links or other references that contain a hijackable domain"

    regex_jssrc = re.compile(r'<script[^>]*src\s*=\s*[\'"]([^\'">]+)[\'"]', re.IGNORECASE)
    regex_csssrc = re.compile(r'<link[^>]*href\s*=\s*[\'"]([^\'">]+)[\'"]', re.IGNORECASE)

    def __init__(self, target, **kwargs):
        super().__init__(target, **kwargs)
        self.target = target
        self.target_dnsmanager = DNSManager(
            target, dns_client=self.dns_client, custom_nameservers=self.custom_nameservers
        )
        self.target_httpmanager = HttpManager(
            self.target, http_client_class=self.http_client_class, skip_redirects=True
        )
        self.cname_findings = None
        self.cname_findings_direct = None
        self.reference_data = {}

    def parse_body(self, body):
        results = []
        for match in self.regex_jssrc.finditer(body):
            js_url = match.group(1)
            parsed_url = urlparse(js_url)
            js_domain = parsed_url.netloc
            results.append(
                {
                    "url": js_url,
                    "domain": js_domain,
                    "description": "Hijackable reference, JS Include",
                    "trigger": f"Javascript Source: [{js_url}]",
                }
            )

        for match in self.regex_csssrc.finditer(body):
            css_url = match.group(1)
            parsed_url = urlparse(css_url)
            css_domain = parsed_url.netloc
            results.append(
                {
                    "url": css_url,
                    "domain": css_domain,
                    "description": "Hijackable reference, CSS Include",
                    "trigger": f"CSS Source: [{css_url}]",
                }
            )
        return results

    async def dispatch(self):
        log.debug("in references dispatch")
        await self.target_httpmanager.dispatchHttp()
        log.debug("HTTP dispatch complete")

        live_results = []

        for protocol in ["http", "https"]:
            result = getattr(self.target_httpmanager, f"{protocol}_denyredirects_results")
            if result:
                log.debug(f"Found live host at {result.url}")
                live_results.append(result)

        self.cname_findings_direct = []
        self.cname_findings = []

        for r in live_results:
            parsed_results = self.parse_body(r.text)
            if parsed_results:
                for pr in parsed_results:
                    if pr["domain"] == self.target:
                        log.debug(f"Found domain matches target ({self.target}), ignoring")
                        continue
                    log.debug(f"Initializing cname instance for target {pr['domain']}")

                    cname_instance_direct = BadDNS_cname(
                        pr["domain"],
                        custom_nameservers=self.custom_nameservers,
                        signatures=self.signatures,
                        direct_mode=True,
                        parent_class="references",
                        http_client_class=self.http_client_class,
                        dns_client=self.dns_client,
                    )
                    if await cname_instance_direct.dispatch():
                        self.cname_findings_direct.append(
                            {
                                "finding": cname_instance_direct.analyze(),
                                "description": pr["description"],
                                "trigger": pr["trigger"],
                            }
                        )

                    cname_instance = BadDNS_cname(
                        pr["domain"],
                        custom_nameservers=self.custom_nameservers,
                        signatures=self.signatures,
                        direct_mode=False,
                        parent_class="references",
                        http_client_class=self.http_client_class,
                        dns_client=self.dns_client,
                    )
                    if await cname_instance.dispatch():
                        self.cname_findings.append(
                            {
                                "finding": cname_instance.analyze(),
                                "description": pr["description"],
                                "trigger": pr["trigger"],
                            }
                        )
        return True

    def _convert_findings(self, finding_sets):
        converted_findings = []
        for finding_set in finding_sets:
            for finding in finding_set["finding"]:
                finding_dict = finding.to_dict()
                log.debug(f"Found finding during cname check for target: {finding_dict['target']}")
                converted_findings.append(
                    Finding(
                        {
                            "target": self.target,
                            "description": f"{finding_set['description']}. Original Event: [{finding_dict['description']}]",
                            "confidence": finding_dict["confidence"],
                            "signature": finding_dict["signature"],
                            "indicator": finding_dict["indicator"],
                            "trigger": finding_set["trigger"],
                            "module": type(self),
                        }
                    )
                )
        return converted_findings

    def analyze(self):
        findings = []
        log.debug("in references analyze")
        if self.cname_findings_direct:
            findings.extend(self._convert_findings(self.cname_findings_direct))
        if self.cname_findings:
            findings.extend(self._convert_findings(self.cname_findings))
        return findings



import os
import importlib
from pathlib import Path
from .base import BadDNS_base

module_dir = Path(__file__).parent / "modules"
module_files = list(os.listdir(module_dir))
modules_loaded = {}
for file in module_files:
    file = module_dir / file
    if file.is_file() and file.suffix.lower() == ".py" and file.stem not in ["base", "__init__"]:
        modules = importlib.import_module(f"baddns.modules.{file.stem}")
        for m in modules.__dict__.keys():
            module = getattr(modules, m)
            try:
                if BadDNS_base in module.__bases__:
                    modules_loaded[file.stem] = module
            except AttributeError:
                continue



import logging
from colorama import Fore, Style

log = None


class CustomLogFormatter(logging.Formatter):
    FORMATS = {
        logging.DEBUG: Fore.MAGENTA + "[%(levelname)s] %(message)s" + Style.RESET_ALL,
        logging.INFO: Fore.CYAN + "%(message)s" + Style.RESET_ALL,
        logging.WARNING: Fore.YELLOW + "[%(levelname)s] %(message)s" + Style.RESET_ALL,
        logging.ERROR: Fore.RED + "[%(levelname)s] %(message)s" + Style.RESET_ALL,
        logging.CRITICAL: Fore.RED + Style.BRIGHT + "[%(levelname)s] - %(message)s" + Style.RESET_ALL,
    }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def setup_logging():
    global log
    log = logging.getLogger("baddns")
    logging.getLogger("httpx").setLevel(logging.WARNING)
    log.setLevel(logging.INFO)
    ch = logging.StreamHandler()
    ch.setFormatter(CustomLogFormatter())
    log.addHandler(ch)


def debug_logging(debug=False):
    global log
    log = logging.getLogger()
    if debug:
        log.setLevel(logging.DEBUG)



import ssl
import anyio
import httpx
import logging

log = logging.getLogger(__name__)


class HttpManager:
    def __init__(self, target, http_client_class=None, skip_redirects=False):
        self.skip_redirects = skip_redirects
        if not http_client_class:
            http_client_class = httpx.AsyncClient
        self.http_client = http_client_class(timeout=5, verify=False)
        self.target = target
        for attr in [
            "http_allowredirects_results",
            "http_denyredirects_results",
            "https_allowredirects_results",
            "https_denyredirects_results",
        ]:
            setattr(self, attr, None)

    async def dispatchHttp(self):
        try:
            protocols = ["http", "https"]
            for protocol in protocols:
                base_url = f"{protocol}://{self.target}/"
                # If redirects are not skipped by this module, perform requests that follow redirects
                if not self.skip_redirects:
                    setattr(
                        self,
                        f"{protocol}_allowredirects_results",
                        await self.http_client.get(base_url, follow_redirects=True),
                    )
                # Always perform requests that do not follow redirects
                setattr(
                    self,
                    f"{protocol}_denyredirects_results",
                    await self.http_client.get(base_url, follow_redirects=False),
                )
        except (httpx.RequestError, httpx.ConnectError, httpx.TimeoutException) as e:
            log.debug(f"An httpx error occurred while requesting {e.request.url!r}: {e}")
        except ssl.SSLError as e:
            log.debug(f"An SSL error occurred while requesting {e}")
        except anyio.EndOfStream as e:
            log.debug(f"An anyio error occurred while requesting {e}")
